(*----------------------------------------------------------------------------*
 *                                                                            *
 *  This unit includes TD3DTextureFont class for rendering text using         *
 *  bitmap based font in Direct3D 8                                           *
 *                                                                            *
 *  Realization by Alexey Barkovoy                                            *
 *  E-Mail: clootie@reactor.ru                                                *
 *                                                                            *
 *  Modified: 10-Nov-2001                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru                                                *
 *----------------------------------------------------------------------------*)

unit D3DTextureFont;

interface

{$I J_D3D.inc}

uses
  Windows, {$I UseD3D8.inc};

type
  // Font rendering flags
  TFontRenderFlag = (rfCentered, rfTwoSided, rfFiltered);
  TFontRenderFlags = set of TFontRenderFlag;

  TD3DTextureFont = class
  private
    m_strFontBitmapFile: String; // Name of bitmap image with font
    m_ColorKey: TD3DColor;       // ColorKey for bitmap with font
    fZEnable: Boolean;           // Do render to Z-buffer

    fWidth, fHeight: DWORD;        // Sizes of Font Texture - used in 2D rendering        

    // DirectX interfaces
    m_pd3dDevice: IDirect3DDevice8;       // A D3DDevice used for rendering
    m_pTexture: IDirect3DTexture8;        // The d3d texture for this font
    m_pVB: IDirect3DVertexBuffer8;        // VertexBuffer for rendering text

    // Stateblocks for setting and restoring render states
    m_dwSavedStateBlock: DWORD;
    m_dwDrawTextStateBlock: DWORD;
    procedure GenerateCharCoords(c: Char; out tx1, ty1, tx2, ty2: Single);
  public
    // 2D and 3D text drawing functions
    function DrawText(x, y: Single; dwColor: DWORD;
                      strText: PChar; Flags: TFontRenderFlags{$IFDEF SUPPORTS_DEFAULTPARAMS} = []{$ENDIF}): HRESULT;
    function DrawTextScaled(x, y, z: Single; fXScale, fYScale: Single; dwColor: DWORD;
                            strText: PChar; Flags: TFontRenderFlags{$IFDEF SUPPORTS_DEFAULTPARAMS} = []{$ENDIF}): HRESULT;
    function Render3DText(strText: PChar; Flags: TFontRenderFlags{$IFDEF SUPPORTS_DEFAULTPARAMS} = []{$ENDIF}): HRESULT;

    // Initializing and destroying device-dependent objects
    function InitDeviceObjects(pd3dDevice: IDirect3DDevice8): HRESULT;
    function RestoreDeviceObjects: HRESULT;
    function InvalidateDeviceObjects: HRESULT;
    function DeleteDeviceObjects: HRESULT;

    // Constructor / destructor
    constructor Create(strFontBitmapFile: String; ColorKey: TD3DColor);
    destructor Destroy; {$IFDEF TMT}virtual{$ELSE}override{$ENDIF};

    property ZEnable: Boolean read fZEnable write fZEnable default True;
    property Width: DWORD read fWidth;
    property Height: DWORD read fHeight;
  end;

implementation

uses
  D3DX8, DXUtil;

//-----------------------------------------------------------------------------
// Custom vertex types for rendering text
//-----------------------------------------------------------------------------
const
  MAX_NUM_VERTICES = 50*6;

type
  PFont2DVertex = ^TFont2DVertex;
  TFont2DVertex = packed record
    p: TD3DXVector4;
    color: DWORD;
    tu, tv: Single;
  end;

  PFont3DVertex = ^TFont3DVertex;
  TFont3DVertex = packed record
    p: TD3DXVector3;
    n: TD3DXVector3;
    tu, tv: Single;
  end;

const
  D3DFVF_Font2DVertex = (D3DFVF_XYZRHW or D3DFVF_DIFFUSE or D3DFVF_TEX1);
  D3DFVF_Font3DVertex = (D3DFVF_XYZ or D3DFVF_NORMAL or D3DFVF_TEX1);

function InitFont2DVertex(const p: TD3DXVector4; color: TD3DColor;
  tu, tv: Single): TFont2DVertex;
begin
  Result.p := p;
  Result.color := color;
  Result.tu := tu;
  Result.tv := tv;
end;

function InitFont3DVertex(const p: TD3DXVector3; const n: TD3DXVector3;
  tu, tv: Single): TFont3DVertex;
begin
  Result.p := p;
  Result.n := n;
  Result.tu := tu;
  Result.tv := tv;
end;


{ TD3DTextureFont }

//-----------------------------------------------------------------------------
// Name: TD3DTextureFont.Create
// Desc: Font class constructor creates object with font from strFontBitmapFile
//       using color ColorKey as transparent (background) color
//-----------------------------------------------------------------------------
constructor TD3DTextureFont.Create(strFontBitmapFile: String; ColorKey: TD3DColor);
begin
  inherited Create;

  m_strFontBitmapFile := strFontBitmapFile;
  m_ColorKey := ColorKey;
  fZEnable:= True;

  m_pd3dDevice          := nil;
  m_pTexture            := nil;
  m_pVB                 := nil;

  m_dwSavedStateBlock    := 0;
  m_dwDrawTextStateBlock := 0;

end;

destructor TD3DTextureFont.Destroy;
begin
  InvalidateDeviceObjects;
  DeleteDeviceObjects;
  inherited Destroy;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initializes device-dependent objects, including the vertex buffer used
//       for rendering text and the texture map which stores the font image.
//-----------------------------------------------------------------------------
function TD3DTextureFont.InitDeviceObjects(
  pd3dDevice: IDirect3DDevice8): HRESULT;
var
  SrcInfo: TD3DXImageInfo;
begin
  // Keep a local copy of the device
  m_pd3dDevice := pd3dDevice;

  // Load texture
  Result:= D3DXCreateTextureFromFileEx(m_pd3dDevice, PChar(m_strFontBitmapFile),
    D3DX_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT, 0, D3DFMT_UNKNOWN, D3DPOOL_MANAGED,
    D3DX_DEFAULT, D3DX_DEFAULT, m_ColorKey, @SrcInfo, nil, m_pTexture);
  if Failed(Result) then Exit;

  // Keep texture dimentions for future 2D font rendering
  fWidth:= SrcInfo.Width;
  fHeight:= SrcInfo.Height;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function TD3DTextureFont.RestoreDeviceObjects: HRESULT;
var
  which: LongWord;
begin
  // Create vertex buffer for the letters
  Result:= m_pd3dDevice.CreateVertexBuffer(MAX_NUM_VERTICES*SizeOf(TFont2DVertex),
                                           D3DUSAGE_WRITEONLY or D3DUSAGE_DYNAMIC,
                                           0, D3DPOOL_DEFAULT, m_pVB);
  if FAILED(Result) then Exit;

  // Create the state blocks for rendering text
  for which:= 0 to 1 do
  begin
    m_pd3dDevice.BeginStateBlock;
    m_pd3dDevice.SetTexture(0, m_pTexture);

    if ZEnable then
      m_pd3dDevice.SetRenderState(D3DRS_ZENABLE, iTRUE)
    else
      m_pd3dDevice.SetRenderState(D3DRS_ZENABLE, iFALSE);

    m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, 1);
    m_pd3dDevice.SetRenderState(D3DRS_SRCBLEND,         D3DBLEND_SRCALPHA);
    m_pd3dDevice.SetRenderState(D3DRS_DESTBLEND,        D3DBLEND_INVSRCALPHA);
    m_pd3dDevice.SetRenderState(D3DRS_ALPHATESTENABLE,  1);
    m_pd3dDevice.SetRenderState(D3DRS_ALPHAREF,         $08);
    m_pd3dDevice.SetRenderState(D3DRS_ALPHAFUNC,        D3DCMP_GREATEREQUAL);
    m_pd3dDevice.SetRenderState(D3DRS_FILLMODE,         D3DFILL_SOLID);
    m_pd3dDevice.SetRenderState(D3DRS_CULLMODE,         D3DCULL_CCW);
    m_pd3dDevice.SetRenderState(D3DRS_STENCILENABLE,    0);
    m_pd3dDevice.SetRenderState(D3DRS_CLIPPING,         1);
    m_pd3dDevice.SetRenderState(D3DRS_EDGEANTIALIAS,    0);
    m_pd3dDevice.SetRenderState(D3DRS_CLIPPLANEENABLE,  0);
    m_pd3dDevice.SetRenderState(D3DRS_VERTEXBLEND,      0);
    m_pd3dDevice.SetRenderState(D3DRS_INDEXEDVERTEXBLENDENABLE, 0);
    m_pd3dDevice.SetRenderState(D3DRS_FOGENABLE,        0);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_MODULATE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_POINT);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_POINT);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTEXF_NONE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_DISABLE);
    m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);

    if (which = 0) then
      m_pd3dDevice.EndStateBlock(m_dwSavedStateBlock)
    else
      m_pd3dDevice.EndStateBlock(m_dwDrawTextStateBlock);
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Destroys all device-dependent objects
//-----------------------------------------------------------------------------
function TD3DTextureFont.InvalidateDeviceObjects: HRESULT;
begin
  SAFE_RELEASE(m_pVB);

  // Delete the state blocks
  if (m_pd3dDevice <> nil) then
  begin
    if (m_dwSavedStateBlock <> 0) then
      m_pd3dDevice.DeleteStateBlock(m_dwSavedStateBlock);
    if (m_dwDrawTextStateBlock <> 0) then
      m_pd3dDevice.DeleteStateBlock(m_dwDrawTextStateBlock);
  end;

  m_dwSavedStateBlock    := 0;
  m_dwDrawTextStateBlock := 0;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Destroys all device-dependent objects
//-----------------------------------------------------------------------------
function TD3DTextureFont.DeleteDeviceObjects: HRESULT;
begin
  SAFE_RELEASE(m_pTexture);
  m_pd3dDevice:= nil; 
  Result:= S_OK;
end;

procedure TD3DTextureFont.GenerateCharCoords(c: Char; out tx1, ty1, tx2, ty2: Single);
var
  LinearEntry, CharX, CharY: DWORD;
begin
  LinearEntry:= 0;
  if (Byte(c) >= 65) and (Byte(c) <= 90) then
    // A character number from 65 through to 90 are the upper case A-Z letters
    // which if we wrap around our texture are entries 0 - 25.
    LinearEntry:= Byte(c) - 65 // Make it so character 65 references entry 0 in our texture
  else if (Byte(c) >= 97) and (Byte(c) <= 122) then
    // We have a lower case letter.
    LinearEntry:= Byte(c) - 71 // Make it so that the lower case letters reference values 26-51
  else if (Byte(c) >= 48) and (Byte(c) <= 57) then
    // We have a numerical character, which occupy entries 52-62 in our texture
    LinearEntry:= Byte(c) + 4

  // Finally: Special Cases. I couldn't be bothered to work out a formula
  //          for full-stop/spaces/punctuation characters, so I'm going to hardcode them
  else
    case c of
      ' ': // Space
           LinearEntry:= 63;
      '.': // Full stop
           LinearEntry:= 62;
      ';': // Semi colon
           LinearEntry:= 66;
      '/': // Forward slash
           LinearEntry:= 64;
      ',': // Guess what; its a comma..
           LinearEntry:= 65;
    end;

  // We now need to process the actual coordinates.
  CharY:= LinearEntry div 16;
  CharX:= LinearEntry mod 16;
  // Fill in the rest if you really need them...

  tx1 := CharX / 16;
  ty1 := CharY / 8;
  tx2 := (CharX + 1) / 16;
  ty2 := (CharY + 1) / 8;
end;

function TD3DTextureFont.DrawText(x, y: Single; dwColor: DWORD;
  strText: PChar; Flags: TFontRenderFlags): HRESULT;
var
  fStartX: Single;
  pVertices: PFont2DVertex;
  dwNumTriangles: DWORD;
  c: Char;
  tx1, ty1, tx2, ty2: Single;
  w, h: Single;
begin
  if (m_pd3dDevice = nil) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  // Set up renderstate
  m_pd3dDevice.CaptureStateBlock(m_dwSavedStateBlock);
  m_pd3dDevice.ApplyStateBlock(m_dwDrawTextStateBlock);
  m_pd3dDevice.SetVertexShader(D3DFVF_FONT2DVERTEX);
  m_pd3dDevice.SetPixelShader(0);
  m_pd3dDevice.SetStreamSource(0, m_pVB, SizeOf(TFont2DVertex));

  // Set filter states
  if (rfFiltered in Flags) then
  begin
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
  end;

  fStartX := x;

  // Fill vertex buffer
  pVertices:= nil;
  dwNumTriangles := 0;
  m_pVB.Lock(0, 0, PByte(pVertices), D3DLOCK_DISCARD);

  while (strText^ <> #0) do
  begin
    c := strText^;
    Inc(strText);

    if (c = #10) then
    begin
      x := fStartX;
      y := y+(1/8);
    end;
    if (c < ' ') then Continue;

    GenerateCharCoords(c, tx1, ty1, tx2, ty2);

    w := (tx2-tx1)*Width;
    h := (ty2-ty1)*Height;

    if (c <> ' ') then
    begin
      pVertices^ := InitFont2DVertex(D3DXVector4(x+0-0.5,y+h-0.5,0.9,1.0), dwColor, tx1, ty2); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(x+0-0.5,y+0-0.5,0.9,1.0), dwColor, tx1, ty1); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(x+w-0.5,y+h-0.5,0.9,1.0), dwColor, tx2, ty2); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(x+w-0.5,y+0-0.5,0.9,1.0), dwColor, tx2, ty1); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(x+w-0.5,y+h-0.5,0.9,1.0), dwColor, tx2, ty2); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(x+0-0.5,y+0-0.5,0.9,1.0), dwColor, tx1, ty1); Inc(pVertices);
      dwNumTriangles :=dwNumTriangles + 2;

      if dwNumTriangles*3 > (MAX_NUM_VERTICES-6)  then
      begin
        // Unlock, render, and relock the vertex buffer
        m_pVB.Unlock;
        m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLELIST, 0, dwNumTriangles);
        pVertices:= nil;
        m_pVB.Lock(0, 0, PByte(pVertices), D3DLOCK_DISCARD);
        dwNumTriangles := 0;
      end;
    end;

    x := x+w;
  end;

  // Unlock and render the vertex buffer
  m_pVB.Unlock;
  if (dwNumTriangles > 0) then
      m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLELIST, 0, dwNumTriangles);

  // Restore the modified renderstates
  m_pd3dDevice.ApplyStateBlock(m_dwSavedStateBlock);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DrawTextScaled()
// Desc: Draws scaled 2D text.  Note that x and y are in viewport coordinates
//       (ranging from -1 to +1).  fXScale and fYScale are the size fraction
//       relative to the entire viewport.  For example, a fXScale of 0.25 is
//       1/8th of the screen width.  This allows you to output text at a fixed
//       fraction of the viewport, even if the screen or window size changes.
//-----------------------------------------------------------------------------
function TD3DTextureFont.DrawTextScaled(x, y, z, fXScale, fYScale: Single;
  dwColor: DWORD; strText: PChar; Flags: TFontRenderFlags): HRESULT;
var
  vp: TD3DViewport8;
  sx, sy, sz, rhw, fStartX, fLineHeight: Single;
  pVertices: PFont2DVertex;
  dwNumTriangles: DWORD;
  c: PChar;
  tx1, ty1, tx2, ty2: Single;
  w, h: Single;
begin
  if (m_pd3dDevice = nil) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  // Set up renderstate
  m_pd3dDevice.CaptureStateBlock(m_dwSavedStateBlock);
  m_pd3dDevice.ApplyStateBlock(m_dwDrawTextStateBlock);
  m_pd3dDevice.SetVertexShader(D3DFVF_FONT2DVERTEX);
  m_pd3dDevice.SetPixelShader(0);
  m_pd3dDevice.SetStreamSource(0, m_pVB, SizeOf(TFont2DVertex));

  // Set filter states
  if (rfFiltered in Flags) then
  begin
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
  end;

  m_pd3dDevice.GetViewport(vp);
  sx  := (x+1.0)*vp.Width/2;
  sy  := (y-1.0)*vp.Height/2;
  sz  := z;
  rhw := 1.0;
  fStartX := sx;

  fLineHeight := Height div 8; // Where are can be 8 lines of letters ...

  // Fill vertex buffer

  dwNumTriangles := 0;
  m_pVB.Lock(0, 0, PByte(pVertices), D3DLOCK_DISCARD);

  while (strText <> nil) do
  begin
    c := strText;
    Inc(strText);

    if (c[0] = #10) then
    begin
      sx := fStartX;
      sy := sy + fYScale*vp.Height;
    end;
    if (c[0] < ' ') then Continue;

    GenerateCharCoords(c[0], tx1, ty1, tx2, ty2);

    w := (tx2-tx1)*Width;
    h := (ty2-ty1)*Height;

    w := w*(fXScale*vp.Height)/fLineHeight;
    h := h*(fYScale*vp.Height)/fLineHeight;

    if (c[0] <> ' ') then
    begin
      pVertices^ := InitFont2DVertex(D3DXVector4(sx+0-0.5, sy+h-0.5, sz, rhw), dwColor, tx1, ty2); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(sx+0-0.5, sy+0-0.5, sz, rhw), dwColor, tx1, ty1); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(sx+w-0.5, sy+h-0.5, sz, rhw), dwColor, tx2, ty2); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(sx+w-0.5, sy+0-0.5, sz, rhw), dwColor, tx2, ty1); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(sx+w-0.5, sy+h-0.5, sz, rhw), dwColor, tx2, ty2); Inc(pVertices);
      pVertices^ := InitFont2DVertex(D3DXVector4(sx+0-0.5, sy+0-0.5, sz, rhw), dwColor, tx1, ty1); Inc(pVertices);
      dwNumTriangles := dwNumTriangles + 2;

      if (dwNumTriangles*3 > (MAX_NUM_VERTICES-6)) then
      begin
        // Unlock, render, and relock the vertex buffer
        m_pVB.Unlock;
        m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLELIST, 0, dwNumTriangles);
        m_pVB.Lock(0, 0, PByte(pVertices), D3DLOCK_DISCARD);
        dwNumTriangles := 0;
      end;
    end;

    sx := sx + w;
  end;

  // Unlock and render the vertex buffer
  m_pVB.Unlock;
  if (dwNumTriangles > 0) then
      m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLELIST, 0, dwNumTriangles);

  // Restore the modified renderstates
  m_pd3dDevice.ApplyStateBlock(m_dwSavedStateBlock);

  Result:= S_OK;
end;

function TD3DTextureFont.Render3DText(strText: PChar; Flags: TFontRenderFlags): HRESULT;
var
  x, y: Single;
  fStartX: Single;
  c: Char;
  pVertices: PFont3DVertex;
  dwNumTriangles: DWORD;
  tx1, ty1, tx2, ty2: Single;
  w, h: Single;
begin
  if (m_pd3dDevice = nil) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  // Setup renderstate
  m_pd3dDevice.CaptureStateBlock(m_dwSavedStateBlock);
  m_pd3dDevice.ApplyStateBlock(m_dwDrawTextStateBlock);
  m_pd3dDevice.SetVertexShader(D3DFVF_FONT3DVERTEX);
  {$IFDEF DX81}
  m_pd3dDevice.SetPixelShader(0);
  {$ENDIF}
  m_pd3dDevice.SetStreamSource(0, m_pVB, SizeOf(TFont3DVertex));

  // Set filter states
  if (rfFiltered in Flags) then
  begin
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
  end;

  // Position for each text element
  x := 0.0;
  y := 0.0;

  // Center the text block at the origin
  if (rfCentered in Flags) then
  begin
    x := -((DWORD(Length(strText))*Width div 16)/10.0)/2.0;
    y := -((Height div 8)/10.0)/2.0;
  end;

  // Turn off culling for two-sided text
  if (rfTwoSided in Flags) then
    m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);

  fStartX := x;

  // Fill vertex buffer
  // dwVertex       := 0;  -  it's not used anyway
  dwNumTriangles := 0;
  m_pVB.Lock(0, 0, PByte(pVertices), D3DLOCK_DISCARD);

  c:= strText^;
  while (c <> #0) do
  begin
    Inc(strText);
    if (c = #10) then
    begin
      x := fStartX;
      y := y-(1/8)*Height/10.0;
    end;
    if (Byte(c) < 32) then Continue;

    GenerateCharCoords(c, tx1, ty1, tx2, ty2);

    w := (tx2-tx1) * Width  / ( 10.0 );
    h := (ty2-ty1) * Height / ( 10.0 );

    if (c <> ' ') then
    begin
      pVertices^ := InitFont3DVertex(D3DXVector3(x+0,y+0,0), D3DXVector3(0,0,-1), tx1, ty2); Inc(pVertices);
      pVertices^ := InitFont3DVertex(D3DXVector3(x+0,y+h,0), D3DXVector3(0,0,-1), tx1, ty1); Inc(pVertices);
      pVertices^ := InitFont3DVertex(D3DXVector3(x+w,y+0,0), D3DXVector3(0,0,-1), tx2, ty2); Inc(pVertices);
      pVertices^ := InitFont3DVertex(D3DXVector3(x+w,y+h,0), D3DXVector3(0,0,-1), tx2, ty1); Inc(pVertices);
      pVertices^ := InitFont3DVertex(D3DXVector3(x+w,y+0,0), D3DXVector3(0,0,-1), tx2, ty2); Inc(pVertices);
      pVertices^ := InitFont3DVertex(D3DXVector3(x+0,y+h,0), D3DXVector3(0,0,-1), tx1, ty1); Inc(pVertices);
      dwNumTriangles := dwNumTriangles + 2;

      if (dwNumTriangles*3 > (MAX_NUM_VERTICES - 6)) then
      begin
        // Unlock, render, and relock the vertex buffer
        m_pVB.Unlock;
        m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLELIST, 0, dwNumTriangles);
        m_pVB.Lock(0, 0, PBYTE(pVertices), D3DLOCK_DISCARD);
        dwNumTriangles := 0;
      end;
    end;

    x := x + w;
    c:= strText^;
  end;

  // Unlock and render the vertex buffer
  m_pVB.Unlock;
  if (dwNumTriangles > 0) then
    m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLELIST, 0, dwNumTriangles);

  // Restore the modified renderstates
  m_pd3dDevice.ApplyStateBlock(m_dwSavedStateBlock);

  Result:= S_OK;
end;

end.

